@@ -2,12 +2,14 @@ from __future__ import absolute_import
 
 import logging
 import os
+import simplejson
 import sys
 import time
 import traceback
 
 from coverage import coverage
 from coverage.report import Reporter
+from collections import defaultdict
 from nose.plugins.base import Plugin
 from nose_bleed.diff import DiffParser
 from operator import or_
@@ -70,7 +72,6 @@ class TestCoverageDB(object):
             statement = select([Coverage.c.lineno, Coverage.c.test], Coverage.c.filename == filename)
             for lineno, test in self.conn.execute(statement).fetchall():
                 file_cover.setdefault(lineno, set()).add(test)
-            self.commit()
         return bool(self._coverage[filename])
 
     def get_test_coverage(self, filename, linenos):
@@ -119,6 +120,7 @@ class TestCoveragePlugin(Plugin):
 
     """
     score = 0
+    name = 'bleed'
 
     def _get_name_from_test(self, test):
         test_method_name = test._testMethodName
@@ -147,6 +149,14 @@ class TestCoveragePlugin(Plugin):
                           dest="record_test_coverage", action="store_false",
                           default=False)
 
+        parser.add_option("--report-test-coverage",
+                          dest="report_test_coverage", action="store_true",
+                          default=False)
+
+        parser.add_option("--no-report-test-coverage",
+                          dest="report_test_coverage", action="store_false",
+                          default=False)
+
         parser.add_option("--skip-missing-coverage",
                           dest="skip_missing_coverage", action="store_true",
                           default=None)
@@ -163,15 +173,40 @@ class TestCoveragePlugin(Plugin):
                           dest="coverage_dsn",
                           default='sqlite:///coverage.db')
 
+        parser.add_option('--with-coverage-file', action='store_true',
+                          default=False)
+
+        parser.add_option('--coverage-file', action='store',
+                          dest='coverage_file', metavar="FILE",
+                          default=env.get('NOSE_BLEED_FILE', 'nosebleed.json'),
+                          help=("Path to json file to store the report in. "
+                                "Default is nosebleed.json in the working directory "
+                                "[NOSE_BLEED_FILE]"))
+
     def configure(self, options, config):
-        self.enabled = (options.record_test_coverage or options.discover)
+        Plugin.configure(self, options, config)
         self.skip_missing = options.skip_missing_coverage
         self.record = options.record_test_coverage
+        self.report_coverage = options.report_test_coverage
         self.discover = options.discover
         self.logger = logging.getLogger(__name__)
         self.dsn = options.coverage_dsn
         self.parent = options.coverage_parent
+        self.enabled = (self.record or self.report_coverage or self.discover)
+
+        if not self.enabled:
+            return
+
         self.pending_funcs = set()
+        # diff is a mapping of filename->[linenos]
+        self.diff_data = defaultdict(set)
+        # cov is a mapping of filename->[linenos]
+        self.cov_data = defaultdict(set)
+
+        if options.with_coverage_file:
+            self.report_file = open(options.coverage_file, 'w')
+        else:
+            self.report_file = None
 
     def begin(self):
         # XXX: this is pretty hacky
@@ -179,7 +214,7 @@ class TestCoveragePlugin(Plugin):
         if self.record:
             self.db.upgrade()
 
-        if not self.discover:
+        if not (self.discover or self.report):
             return
 
         s = time.time()
@@ -193,20 +228,27 @@ class TestCoveragePlugin(Plugin):
 
         parser = DiffParser(diff)
         files = list(parser.parse())
-        self.logger.info("Parsed diff in %.2fs", time.time() - s)
 
-        self.logger.info("Finding coverage for %d file(s)", len(files))
+        diff = self.diff_data
         s = time.time()
         for file in files:
+            # we dont care about headers
             if file['is_header']:
                 continue
 
-            if file['old_filename'] == '/dev/null':
+            # file was removed
+            if file['new_filename'] == '/dev/null':
                 continue
 
-            filename = file['old_filename']
-            if not filename.startswith('a/'):
-                continue # ??
+            is_new_file = (file['old_filename'] == '/dev/null')
+            if is_new_file:
+                filename = file['new_filename']
+                if not filename.startswith('b/'):
+                    continue
+            else:
+                filename = file['old_filename']
+                if not filename.startswith('a/'):
+                    continue # ??
 
             filename = filename[2:]
 
@@ -214,26 +256,85 @@ class TestCoveragePlugin(Plugin):
             if not is_py_script(filename):
                 continue
 
-            # TODO: for this to be useful we need to eliminate tests
-            if not self.db.has_test_coverage(filename):
-                if self.skip_missing:
-                    self.logger.warning('%s has no test coverage recorded', filename)
-                    continue
-                raise AssertionError("Missing test coverage for %s" % filename)
-
+            # file is new, only record diff state
             for chunk in file['chunks']:
-                linenos = [l['old_lineno'] for l in chunk]
+                linenos = filter(bool, (l['new_lineno'] for l in chunk))
+                diff[file['new_filename'][2:]].update(linenos)
+
+            # we dont care about missing coverage for new code, and there
+            # wont be any "existing coverage" to check for
+            if is_new_file:
+                continue
+
+            if not self.discover:
+                continue
+
+        self.logger.info("Parsed diff in %.2fs", time.time() - s)
+
+        if self.discover:
+            self.logger.info("Finding coverage for %d file(s)", len(files))
+
+            for filename, linenos in diff.iteritems():
+                # TODO: for this to be useful we need to eliminate tests
+                if not self.db.has_test_coverage(filename):
+                    if self.skip_missing:
+                        self.logger.warning('%s has no test coverage recorded', filename)
+                        continue
+                    raise AssertionError("Missing test coverage for %s" % filename)
+
+                # record
                 for test in self.db.get_test_coverage(filename, linenos):
                     pending_funcs.add(test)
 
-        self.logger.info("Determined available coverage in %.2fs with %d test(s)", time.time() - s, len(pending_funcs))
+            self.logger.info("Determined available coverage in %.2fs with %d test(s)", time.time() - s, len(pending_funcs))
+
+    def report(self, stream):
+        if not self.report_coverage:
+            return
+
+        covered = 0
+        total = 0
+        missing = defaultdict(set)
+        for filename, linenos in self.diff_data.iteritems():
+            covered_linenos = self.cov_data[filename]
+
+            total += len(linenos)
+            covered += len(covered_linenos)
+
+            missing[filename] = linenos.difference(covered_linenos)
+
+
+        if self.report_file:
+            self.report_file.write(simplejson.dumps({
+                'stats': {
+                    'covered': covered,
+                    'total': total,
+                },
+                'missing': dict((k, tuple(v)) for k, v in missing.iteritems() if v),
+            }))
+            self.report_file.close()
+        else:
+            stream.writeln('Coverage Report')
+            stream.writeln('-'*70)
+            stream.writeln('Coverage against diff is %.2f%% (%d / %d lines)' % (covered / float(total) * 100, covered, total))
+            if missing:
+                stream.writeln()
+                stream.writeln('%-35s   %s' % ('Filename', 'Missing Lines'))
+                stream.writeln('-'*70)
+                for filename, linenos in missing.iteritems():
+                    if not linenos:
+                        continue
+                    stream.writeln('%-35s   %s' % (filename, ', '.join(map(str, sorted(linenos)))))
+
 
     def wantMethod(self, method):
         if not self.discover:
             return
 
         # only works with unittest compatible functions currently
-        test_name = '%s:%s.%s' % (method.im_class.__module__, method.im_class.__name__, method.__name__)
+        method_name = method.__name__
+        method = getattr(sys.modules[method.im_class.__module__], method.im_class.__name__)
+        test_name = '%s:%s.%s' % (method.__module__, method.__name__, method_name)
 
         # test has coverage for diff
         if test_name in self.pending_funcs:
@@ -248,14 +349,14 @@ class TestCoveragePlugin(Plugin):
         return False
 
     def startTest(self, test):
-        if not self.record:
+        if not (self.report_coverage or self.record):
             return
 
         self.coverage = coverage(include='disqus/*')
         self.coverage.start()
 
     def stopTest(self, test):
-        if not self.record:
+        if not (self.report_coverage or self.record):
             return
 
         cov = self.coverage
@@ -265,9 +366,9 @@ class TestCoveragePlugin(Plugin):
         test_name = self._get_name_from_test(test_)
 
         # this must have been imported under a different name
-        if self.discover and test_name not in self.pending_funcs:
-            self.logger.warning("Unable to determine origin for test: %s", test_name)
-            return
+        # if self.discover and test_name not in self.pending_funcs:
+        #     self.logger.warning("Unable to determine origin for test: %s", test_name)
+        #     return
 
         # initialize reporter
         rep = Reporter(cov)
@@ -278,16 +379,24 @@ class TestCoveragePlugin(Plugin):
         # The rest of this is all run within a single transaction
         trans = self.db.begin()
 
-        self.db.clear_test_coverage(test_name)
+        if self.record:
+            self.db.clear_test_coverage(test_name)
 
         for cu in rep.code_units:
-            if sys.modules[test_.__module__].__file__ == cu.filename:
-                continue
+            # if sys.modules[test_.__module__].__file__ == cu.filename:
+            #     continue
             filename = cu.name + '.py'
             try:
                 # TODO: this CANT work in all cases, must be a better way
                 analysis = rep.coverage._analyze(cu)
-                self.db.set_test_coverage(test_name, filename, analysis.statements)
+                linenos = analysis.statements
+                if self.record:
+                    self.db.set_test_coverage(test_name, filename, linenos)
+                if self.report_coverage:
+                    diff = self.diff_data[filename]
+                    cov_linenos = [l for l in linenos if l in diff]
+                    if cov_linenos:
+                        self.cov_data[filename].update(cov_linenos)
             except KeyboardInterrupt:                       # pragma: no cover
                 raise
             except:
